#include "gglobal.h"
#include "gevent.h"
#include "gideros.h"
#include "gapplication.h"

// some Lua helper functions
#ifndef abs_index
#define abs_index(L, i) ((i) > 0 || (i) <= LUA_REGISTRYINDEX ? (i) : lua_gettop(L) + (i) + 1)
#endif

static char keyWeak = ' ';
static lua_State *L = NULL;

static void luaL_newweaktable(lua_State *L, const char *mode)
{
    lua_newtable(L);			// create table for instance list
    lua_pushstring(L, mode);
    lua_setfield(L, -2, "__mode");	  // set as weak-value table
    lua_pushvalue(L, -1);             // duplicate table
    lua_setmetatable(L, -2);          // set itself as metatable
}

static void luaL_rawgetptr(lua_State *L, int idx, void *ptr)
{
    idx = abs_index(L, idx);
    lua_pushlightuserdata(L, ptr);
    lua_rawget(L, idx);
}

static void luaL_rawsetptr(lua_State *L, int idx, void *ptr)
{
    idx = abs_index(L, idx);
    lua_pushlightuserdata(L, ptr);
    lua_insert(L, -2);
    lua_rawset(L, idx);
}

typedef struct gopenurl_event
{
    const char *url;
} gopenurl_event;

static const char *OPEN_URL = "openUrl";

class GOpenUrl : public GEventDispatcherProxy
{
public:
    GOpenUrl()
    {
        gid_ = g_NextId();
        gapplication_addCallback(openUrl_s, this);
    }

    ~GOpenUrl()
    {
        gapplication_removeCallback(openUrl_s, this);
        gevent_RemoveEventsWithGid(gid_);
    }
private:
    static void openUrl_s(int type, void *event, void *udata)
    {
        static_cast<GOpenUrl*>(udata)->openUrl(type, event);
    }

    void openUrl(int type, void *event)
    {

        if (type == GAPPLICATION_OPEN_URL_EVENT)
        {
            gapplication_OpenUrlEvent *event2 = (gapplication_OpenUrlEvent*)event;
            const char* url = event2->url;

            gopenurl_event *eventRet = (gopenurl_event*)gevent_CreateEventStruct1(
                sizeof(gopenurl_event),
                offsetof(gopenurl_event, url), url);

            gevent_EnqueueEvent(gid_, callback_s, 0, eventRet, 1, this);
        }
    }

private:
    static void callback_s(int type, void *event, void *udata)
    {
        static_cast<GOpenUrl*>(udata)->callback(type, event);
    }

    void callback(int type, void *event)
    {
        dispatchEvent(type, event);
    }

    void dispatchEvent(int type, void *event)
    {
        if (L == NULL)
            return;

        luaL_rawgetptr(L, LUA_REGISTRYINDEX, &keyWeak);
        luaL_rawgetptr(L, -1, this);

        if (lua_isnil(L, -1))
        {
            lua_pop(L, 2);
            return;
        }

        lua_getfield(L, -1, "dispatchEvent");

        lua_pushvalue(L, -2);

        lua_getglobal(L, "Event");
        lua_getfield(L, -1, "new");
        lua_remove(L, -2);

        lua_pushstring(L, OPEN_URL);

        lua_call(L, 1, 1);

        gopenurl_event *event2 = (gopenurl_event*)event;

        lua_pushstring(L, event2->url);
        lua_setfield(L, -2, "url");

        lua_call(L, 2, 0);

        lua_pop(L, 2);
    }

private:
    gevent_CallbackList callbackList_;
    g_id gid_;
};


static int destruct(lua_State* L)
{
    void *ptr = *(void**)lua_touserdata(L, 1);
    GReferenced* object = static_cast<GReferenced*>(ptr);
    GOpenUrl *inst = static_cast<GOpenUrl*>(object->proxy());

    inst->unref();

    return 0;
}

static int luaopen_openurl(lua_State *L)
{
    const luaL_Reg functionlist[] = {
        {NULL, NULL},
    };
    g_createClass(L, "OpenUrl", "EventDispatcher", NULL, destruct, functionlist);
    GOpenUrl *inst = new GOpenUrl;
    g_pushInstance(L, "OpenUrl", inst->object());

    luaL_newweaktable(L, "v");
    luaL_rawsetptr(L, LUA_REGISTRYINDEX, &keyWeak);

    lua_getglobal(L, "Event");
    lua_pushstring(L, OPEN_URL);
    lua_setfield(L, -2, "OPEN_URL");
    lua_pop(L, 1);

    luaL_rawgetptr(L, LUA_REGISTRYINDEX, &keyWeak);
    lua_pushvalue(L, -2);
    luaL_rawsetptr(L, -2, inst);
    lua_pop(L, 1);

    lua_pushvalue(L, -1);
    lua_setglobal(L, "openurl");
    return 1;
}

static void g_initializePlugin(lua_State *L)
{
    ::L = L;

    lua_getglobal(L, "package");
    lua_getfield(L, -1, "preload");

    lua_pushcfunction(L, luaopen_openurl);
    lua_setfield(L, -2, "openurl");

    lua_pop(L, 2);
}

static void g_deinitializePlugin(lua_State *L)
{
    ::L = NULL;
}

REGISTER_PLUGIN("OpenUrl", "1.0")

